import asyncio

class TaskPool(object):

    def __init__(self, workers):
        self._semaphore = asyncio.Semaphore(workers)
        self._tasks = set()
        self._closed = False
        

    async def put(self, coro):

        if self._closed:
            raise RuntimeError("Trying put items into a closed TaskPool")
        
        await self._semaphore.acquire()
        
        task = asyncio.ensure_future(coro)
        self._tasks.add(task)
        task.add_done_callback(self._on_task_done)

    def _on_task_done(self, task):
        self._tasks.remove(task)
        self._semaphore.release()

    async def join(self):
        await asyncio.gather(*self._tasks)
        self._closed = True

    async def __aenter__(self):
        return self

    def __aexit__(self, exc_type, exc, tb):
        return self.join()    


"""
Hola David
"""

from __future__ import absolute_import, print_function



import sys
from . import pr
from . import th

if sys.version_info >= (3, 5):
    from . import io
    from .task_pool import TaskPool

__all__ = ["pr", "th", "io"]


import functools
import traceback
from collections import namedtuple

try:
    from wrapt import decorator as wrapt_decorator
except ImportError:
    def wrapt_decorator(f):

        @functools.wraps(f)
        def wrapper_f(g):

            @functools.wraps(g)
            def wrapper_g(*args, **kwargs):
                return f(g, None, args, kwargs)

            return wrapper_g

        return wrapper_f



TIMEOUT = 0.0001

class Partial(object):

    def __init__(self, f):
        self.f = f

    def __or__(self, stage):
        return self.f(stage)

    def __ror__(self, stage):
        return self.f(stage)

    def __call__(self, stage):
        return self.f(stage)

class BaseStage(object):

    def __or__(self, f):
        return f(self)

    

class Namespace(object):
    pass

class Done(object): pass
DONE = Done()
def is_done(x): return isinstance(x, Done)

class Value(namedtuple("Value", "value")): pass
def is_value(x): return isinstance(x, Value)

class Continue(object): pass
CONTINUE = Continue()
def is_continue(x): return isinstance(x, Continue)

class _None(object): pass
NONE = _None()
def is_none(x): return isinstance(x, _None)


def chunks(n, l):
    """Yield successive n-sized chunks from l."""
    for i in range(0, len(l), n):
        if i + n <= len(l):
            yield l[i:i + n]



    
def maybe_partial(n):

    @wrapt_decorator
    def wrapper(wrapped_f, instance, args, kwargs):

        if len(args) < n:
            return Partial(lambda s: wrapped_f(*(args + (s,)), **kwargs))
        else:
            return wrapped_f(*args, **kwargs)

    return wrapper


def print_error(f):

    @functools.wraps(f)
    def _lambda(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception as e:
            print(traceback.format_exc())
            raise e

    return _lambda



import os
from setuptools import setup, find_packages

install_requires = [

]

setup(
    name = "pypeln",
    version = "0.0.11",
    author = "Cristian Garcia",
    author_email = "cgarcia.e88@gmail.com",
    description = (""),
    license = "MIT",
    keywords = [],
    url = "https://github.com/cgarciae/pypeln",
   	packages = find_packages(),
    package_data={
        '': ['LICENCE', 'requirements.txt', 'README.md', 'CHANGELOG.md'],
    },
    include_package_data = True,
    install_requires = install_requires,
)



from __future__ import absolute_import, print_function

from collections import namedtuple
import asyncio
import threading
import sys
from . import utils
from .task_pool import TaskPool
import time
from functools import reduce

def WORKER(target, args, kwargs):
    return target(*args, **kwargs)

class Stage(utils.BaseStage):

    def __init__(self, worker_constructor, workers, maxsize, target, args, dependencies):
        self.worker_constructor = worker_constructor
        self.workers = workers
        self.maxsize = maxsize
        self.target = target
        self.args = args
        self.dependencies = dependencies

    def __await__(self):
        task, _input_queue = _to_task(self, maxsize = 0)
        return task.__await__()
    
    def __iter__(self):
        return to_iterable(self)

    def __repr__(self):
        return "Stage(worker_constructor = {worker_constructor}, workers = {workers}, maxsize = {maxsize}, target = {target}, args = {args}, dependencies = {dependencies})".format(
            worker_constructor = self.worker_constructor,
            workers = self.workers,
            maxsize = self.maxsize,
            target = self.target,
            args = self.args,
            dependencies = len(self.dependencies),
        )

class InputQueue(asyncio.Queue):

    def __init__(self, maxsize = 0, total_done = 1, **kwargs):
        super(InputQueue, self).__init__(maxsize = maxsize, **kwargs)

        self.remaining = total_done

    async def __aiter__(self):

        while not self.is_done():
            x = await self.get()

            if not utils.is_continue(x):
                yield x

    async def get(self):
        
        x = await super(InputQueue, self).get()

        if utils.is_done(x):
            self.remaining -= 1    
            return utils.CONTINUE

        return x

    def get_nowait(self):
        
        x = super(InputQueue, self).get_nowait()

        if utils.is_done(x):
            self.remaining -= 1    
            return utils.CONTINUE

        return x
            

    def is_done(self):
        return self.remaining == 0 # and self.empty()


class OutputQueues(list):

    async def put(self, x):
        for queue in self:
            await queue.put(x)

    async def done(self):
        for queue in self:
            await queue.put(utils.DONE)

    async def __aenter__(self):
        return self

    def __aexit__(self, exc_type, exc, tb):
        return self.done()


async def _run_tasks(f_task, workers, input_queue, output_queues):

    async with TaskPool(workers = workers) as tasks:
        async for x in input_queue:
            
            task = f_task(x)
            await tasks.put(task)

    # wait all tasks to finish
    await output_queues.done()

########################
# map
########################

async def _map(f, workers, input_queue, output_queues):

    async def f_task(x):

        y = f(x)

        if hasattr(y, "__await__"):
            y = await y

        await output_queues.put(y)

    await _run_tasks(f_task, workers, input_queue, output_queues)


@utils.maybe_partial(2)
def map(f, stage = None, workers = 1, maxsize = 0):

    stage = _to_stage(stage)

    return Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _map,
        args = (f,),
        dependencies = [stage],
    )



########################
# flat_map
########################

async def _flat_map(f, workers, input_queue, output_queues):
    

    async def f_task(x):

        ys = f(x)

        if hasattr(ys, "__aiter__"):
            async for y in ys:
                await output_queues.put(y)
            
        elif hasattr(ys, "__iter__"):
            for y in ys:
                await output_queues.put(y)

    await _run_tasks(f_task, workers, input_queue, output_queues)


@utils.maybe_partial(2)
def flat_map(f, stage = None, workers = 1, maxsize = 0):

    stage = _to_stage(stage)

    return Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _flat_map,
        args = (f,),
        dependencies = [stage],
    )

########################
# filter
########################

async def _filter(f, workers, input_queue, output_queues):
    
    async def f_task(x):

        y = f(x)

        if hasattr(y, "__await__"):
            y = await y

        if y:
            await output_queues.put(x)

    await _run_tasks(f_task, workers, input_queue, output_queues)


@utils.maybe_partial(2)
def filter(f, stage = None, workers = 1, maxsize = 0):

    stage = _to_stage(stage)

    return Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _filter,
        args = (f,),
        dependencies = [stage],
    )


########################
# each
########################


async def _each(f, workers, input_queue, output_queues):
    
    async def f_task(x):

        y = f(x)

        if hasattr(y, "__await__"):
            y = await y

    await _run_tasks(f_task, workers, input_queue, output_queues)


@utils.maybe_partial(2)
def each(f, stage = None, workers = 1, maxsize = 0, run = True):

    stage = _to_stage(stage)

    stage = Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _each,
        args = (f,),
        dependencies = [stage],
    )

    if not run:
        return stage

    for _ in stage:
        pass

########################
# concat
########################

async def _concat(workers, input_queue, output_queues):

    async for x in input_queue:
        await output_queues.put(x)
        
    await output_queues.done()
    


def concat(stages, maxsize = 0):

    stages = [ _to_stage(s) for s in stages ]

    return Stage(
        worker_constructor = WORKER,
        workers = 1,
        maxsize = maxsize,
        target = _concat,
        args = tuple(),
        dependencies = stages,
    )


################
# _to_task
################

def _to_task(stage, maxsize):

    total_done = 1
    input_queue = InputQueue(maxsize, total_done)

    stage_input_queue, stage_output_queues = _build_queues(
        stage = stage,
        stage_input_queue = dict(),
        stage_output_queues = dict(),
        visited = set(),
    )

    stage_output_queues[stage] = OutputQueues([ input_queue ])

    tasks = [
        _stage.worker_constructor(
            target = _stage.target,
            args = _stage.args,
            kwargs = dict(
                workers = _stage.workers,
                output_queues = stage_output_queues[_stage],
                input_queue = stage_input_queue.get(_stage, None),
            ),
        )
        for _stage in stage_output_queues
    ]

    task = asyncio.gather(*tasks)

    return task, input_queue

################
# _to_stage
################

def _to_stage(obj):

    if isinstance(obj, Stage):
        return obj

    elif hasattr(obj, "__iter__"):
        return from_iterable(obj)

    else:
        raise ValueError("Object {obj} is not iterable".format(obj = obj))
    
########################
# from_iterable
########################

async def _from_iterable(iterable, workers, input_queue, output_queues):

    for x in iterable:
        await output_queues.put(x)
        
    await output_queues.done()

@utils.maybe_partial(1)
def from_iterable(iterable = None, worker_constructor = None):
    
    return Stage(
        worker_constructor = WORKER,
        workers = 1,
        maxsize = None,
        target = _from_iterable,
        args = (iterable,),
        dependencies = [],
    )



########################
# to_iterable
########################


def _build_queues(stage, stage_input_queue, stage_output_queues, visited):

    if stage in visited:
        return stage_input_queue, stage_output_queues
    else:
        visited.add(stage)

    if len(stage.dependencies) > 0:
        total_done = len(stage.dependencies)
        input_queue = InputQueue(stage.maxsize, total_done)
        stage_input_queue[stage] = input_queue

        for stage in stage.dependencies:
            if stage not in stage_output_queues:
                stage_output_queues[stage] = OutputQueues([input_queue])
            else:
                stage_output_queues[stage].append(input_queue)

            stage_input_queue, stage_output_queues = _build_queues(
                stage,
                stage_input_queue,
                stage_output_queues,
                visited
            )

    return stage_input_queue, stage_output_queues

def _handle_async_exception(loop, ctx, error_namespace, task):
    print("CTX", ctx)
    if error_namespace.exception is None:
        if "exception" in ctx:
            error_namespace.exception = ctx["exception"]
        if "message" in ctx:
            error_namespace.exception = Exception(ctx["message"])
        else:
            error_namespace.exception = Exception(str(ctx))
        
        task.cancel()

def _to_iterable_fn(loop, task):
    
    loop.run_until_complete(task)


@utils.maybe_partial(1)
def to_iterable(stage: Stage = None, maxsize = 0):

    error_namespace = utils.Namespace()
    error_namespace.exception = None

    # old_loop = asyncio.get_event_loop()
    loop = asyncio.get_event_loop()
    # loop = asyncio.new_event_loop()
    # asyncio.set_event_loop(loop)
    

    task, input_queue = _to_task(stage, maxsize = maxsize)

    # exception handler
    old_exception_handler = loop.get_exception_handler()

    def exception_handler(loop, ctx):
        if old_exception_handler:
            old_exception_handler(loop, ctx)
        _handle_async_exception(loop, ctx, error_namespace, task)
    
    loop.set_exception_handler(exception_handler)

    # start thread
    thread = threading.Thread(target=_to_iterable_fn, args=(loop, task))
    thread.daemon = True
    thread.start()

    try:
        while not input_queue.is_done():

            while input_queue.empty():
                time.sleep(utils.TIMEOUT)

            x = input_queue.get_nowait()

            if error_namespace.exception is not None:
                raise error_namespace.exception

            if not utils.is_continue(x):
                yield x
        
        thread.join()
        
    finally:
        # asyncio.set_event_loop(old_loop)
        loop.set_exception_handler(old_exception_handler)
        pass




if __name__ == '__main__':
    import random

    async def slow_square(x):
        await asyncio.sleep(random.uniform(0, 1))
        return x**2

    stage = range(10)

    stage = flat_map(lambda x: [x, x + 1, x + 2], stage)

    stage = map(slow_square, stage, workers = 4)

    stage = filter(lambda x: x > 9, stage)

    each(print, stage)


    



from __future__ import absolute_import, print_function

from functools import reduce
from collections import namedtuple
from . import utils

#############
# imports pr
#############

from multiprocessing import Process as WORKER
from multiprocessing import Manager, Lock, Queue
from multiprocessing.queues import Full, Empty
from threading import Thread

from collections import namedtuple
from . import utils

MANAGER = Manager()

def _get_namespace():
    return MANAGER.Namespace()

#############
# imports th
#############

# from threading import Thread as WORKER
# from threading import Thread
# from.utils import Namespace
# from six.moves.queue import Queue, Empty, Full
# from threading import Lock

# def _get_namespace():
#     return Namespace()


####################
# classes
####################

class Stage(utils.BaseStage):

    def __init__(self, worker_constructor, workers, maxsize, target, args, dependencies):
        self.worker_constructor = worker_constructor
        self.workers = workers
        self.maxsize = maxsize
        self.target = target
        self.args = args
        self.dependencies = dependencies
    
    def __iter__(self):
        return to_iterable(self)

    def __repr__(self):
        return "Stage(worker_constructor = {worker_constructor}, workers = {workers}, maxsize = {maxsize}, target = {target}, args = {args}, dependencies = {dependencies})".format(
            worker_constructor = self.worker_constructor,
            workers = self.workers,
            maxsize = self.maxsize,
            target = self.target,
            args = self.args,
            dependencies = len(self.dependencies),
        )


class InputQueue(object):

    def __init__(self, maxsize, total_done, **kwargs):
        
        self.queue = Queue(maxsize = maxsize, **kwargs)
        self.lock = Lock()
        self.namespace = _get_namespace()
        self.namespace.remaining = total_done

    def __iter__(self):

        while not self.is_done():
            x = self.get()

            if not utils.is_continue(x):
                yield x

    def get(self):
        
        try:
            x = self.queue.get(timeout = utils.TIMEOUT)
        except (Empty, Full):
            return utils.CONTINUE
        
        if not utils.is_done(x):
            return x
        else:
            with self.lock:
                self.namespace.remaining -= 1
            
            return utils.CONTINUE

    def is_done(self):
        return self.namespace.remaining == 0 and self.queue.empty()

    def put(self, x):
        self.queue.put(x)


class OutputQueues(list):

    def put(self, x):
        for queue in self:
            queue.put(x)

    def done(self):
        for queue in self:
            queue.put(utils.DONE)

###########
# map
###########

def _map(f, input_queue, output_queues):

    for x in input_queue:
        y = f(x)
        output_queues.put(y)


    output_queues.done()


@utils.maybe_partial(2)
def map(f, stage = None, workers = 1, maxsize = 0):

    stage = _to_stage(stage)

    return Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _map,
        args = (f,),
        dependencies = [stage],
    )

###########
# flat_map
###########

def _flat_map(f, input_queue, output_queues):

    for x in input_queue:
        for y in f(x):
            output_queues.put(y)

    output_queues.done()


@utils.maybe_partial(2)
def flat_map(f, stage = None, workers = 1, maxsize = 0):

    stage = _to_stage(stage)

    return Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _flat_map,
        args = (f,),
        dependencies = [stage],
    )


###########
# filter
###########

def _filter(f, input_queue, output_queues):

    for x in input_queue:
        if f(x):
            output_queues.put(x)

    output_queues.done()


@utils.maybe_partial(2)
def filter(f, stage = None, workers = 1, maxsize = 0):

    stage = _to_stage(stage)

    return Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _filter,
        args = (f,),
        dependencies = [stage],
    )


###########
# each
###########

def _each(f, input_queue, output_queues):

    for x in input_queue:
        f(x)

    output_queues.done()


@utils.maybe_partial(2)
def each(f, stage = None, workers = 1, maxsize = 0, run = True):

    stage = _to_stage(stage)

    stage = Stage(
        worker_constructor = WORKER,
        workers = workers,
        maxsize = maxsize,
        target = _each,
        args = (f,),
        dependencies = [stage],
    )

    if not run:
        return stage

    for _ in stage:
        pass


###########
# concat
###########

def _concat(input_queue, output_queues):

    for x in input_queue:
        output_queues.put(x)

    output_queues.done()


def concat(stages, maxsize = 0):

    stages = [ _to_stage(s) for s in stages ]

    return Stage(
        worker_constructor = WORKER,
        workers = 1,
        maxsize = maxsize,
        target = _concat,
        args = tuple(),
        dependencies = stages,
    )

################
# run
################

def run(stages, maxsize = 0):
    
    if isinstance(stages, list) and len(stages) == 0:
        raise ValueError("Expected atleast stage to run")

    elif isinstance(stages, list):
        stage = concat(stages, maxsize = maxsize)
    
    else:
        stage = stages

    stage = to_iterable(stage, maxsize = maxsize)
    
    for _ in stages:
        pass

    

################
# _to_stage
################ 

def _to_stage(obj):

    if isinstance(obj, Stage):
        return obj

    elif hasattr(obj, "__iter__"):
        return from_iterable(obj)
    
    else:
        raise ValueError("Object {obj} is not iterable".format(obj = obj))

################
# from_iterable
################

def _from_iterable(iterable, input_queue, output_queues):

    for x in iterable:
        output_queues.put(x)
    
    output_queues.done()

@utils.maybe_partial(1)
def from_iterable(iterable = None, worker_constructor = Thread):

    return Stage(
        worker_constructor = worker_constructor,
        workers = 1,
        maxsize = None,
        target = _from_iterable,
        args = (iterable,),
        dependencies = [],
    )

##############
# to_iterable
##############

def _build_queues(stage, stage_input_queue, stage_output_queues, visited):

    if stage in visited:
        return stage_input_queue, stage_output_queues
    else:
        visited.add(stage)
    
    
    if len(stage.dependencies) > 0:
        total_done = sum([ s.workers for s in stage.dependencies ])
        input_queue = InputQueue(stage.maxsize, total_done)
        stage_input_queue[stage] = input_queue

        for _stage in stage.dependencies:
            
            if _stage not in stage_output_queues:
                stage_output_queues[_stage] = OutputQueues([input_queue])
            else:
                stage_output_queues[_stage].append(input_queue)

            stage_input_queue, stage_output_queues = _build_queues(
                _stage,
                stage_input_queue,
                stage_output_queues,
                visited
            )

    return stage_input_queue, stage_output_queues

def _create_worker(f, args, output_queues, input_queue):

    kwargs = dict(
        output_queues = output_queues)

    if input_queue is not None:
        kwargs.update(input_queue = input_queue)

    return WORKER(target = f, args = args, kwargs = kwargs)

@utils.maybe_partial(1)
def to_iterable(stage = None, maxsize = 0):

    input_queue = InputQueue(maxsize, stage.workers)

    stage_input_queue, stage_output_queues = _build_queues(
        stage = stage,
        stage_input_queue = dict(),
        stage_output_queues = dict(),
        visited = set(),
    )

    stage_output_queues[stage] = OutputQueues([ input_queue ])

    processes = [
        _stage.worker_constructor(
            target = _stage.target,
            args = _stage.args,
            kwargs = dict(
                output_queues = stage_output_queues[_stage],
                input_queue = stage_input_queue.get(_stage, None),
            ),
        )
        for _stage in stage_output_queues
        for _ in range(_stage.workers)
    ]

    for p in processes:
        p.daemon = True
        p.start()

    for x in input_queue:
        yield x

    
    for p in processes:
        p.join()

if __name__ == '__main__':
    import time
    import random

    def slow_square(x):
        time.sleep(random.uniform(0, 1))
        return x**2

    stage = range(10)

    stage = flat_map(lambda x: [x, x + 1, x + 2], stage)

    stage = map(slow_square, stage, workers=4)

    stage = filter(lambda x: x > 9, stage)

    print(stage)
    

    
